/**
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 * 
 * 
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * a) This file is available under the Apache License, Version 2.0.
 * b) When publishing any derivative work or results obtained using this software, you agree to cite the following paper:
 *    Tomas Koutny and Martin Ubl, "SmartCGMS as a Testbed for a Blood-Glucose Level Prediction and/or 
 *    Control Challenge with (an FDA-Accepted) Diabetic Patient Simulation", Procedia Computer Science,  
 *    Volume 177, pp. 354-362, 2020
 */

#include "generated_model.h"
#include <scgms/rtl/SolverLib.h>
#include <scgms/rtl/rattime.h>

#include <type_traits>
#include <cassert>

/**********************************
 * Generated model implementation *
 **********************************/

CGenerated123_Model::CGenerated123_Model(scgms::IModel_Parameter_Vector *parameters, scgms::IFilter *output) :
	CBase_Filter(output),
	// BP NOTE: jen generujte spravne jmeno namespace
	mParameters(scgms::Convert_Parameters<generated123_model::TParameters>(parameters, generated123_model::default_parameters.vector)),

	// BP NOTE: opakujte pro kazdy kompartment dle vzoru
	mEquation_Binding{
		{ mState.C[0], std::bind<double>(&CGenerated123_Model::eq_dC0, this, std::placeholders::_1, std::placeholders::_2) },
		{ mState.C[1], std::bind<double>(&CGenerated123_Model::eq_dC1, this, std::placeholders::_1, std::placeholders::_2) },
		{ mState.C[2], std::bind<double>(&CGenerated123_Model::eq_dC2, this, std::placeholders::_1, std::placeholders::_2) },
		{ mState.C[3], std::bind<double>(&CGenerated123_Model::eq_dC3, this, std::placeholders::_1, std::placeholders::_2) },
	}
{
	mState.lastTime = -std::numeric_limits<decltype(mState.lastTime)>::max();
	mInitialized = false;

	for (size_t i = 0; i < Num_Compartments; i++) {
		mState.C[i] = 0;
	}
}

// BP NOTE: nasledujici metody vygenerujte podle toho, jak uzivatel propojil kompartmenty

/*

Priklad zde:

A) Kompartment 0 je spojeny s kompartmentem 1 exponencialnim prenosem s parametrem 2 (C0 --(p2)--> C1)
B) Kompartment 0 je spojeny s kompartmentem 2 exponencialnim prenosem s parametrem 3 (C0 --(p3)--> C2)
C) Kompartment 1 je spojeny s kompartmentem 3 difuznim prenosem s parametrem 0 (C1 <--(p0)--> C3)            -- pozn. difuzni vazba = mnozstvi odchazi z kompartmentu, kde je ho vic do toho, kde je ho mene
D) Z kompartmentu 2 dochazi k exponencialni eliminaci s parametrem 1 (C2 --(p1)--> _)
E) V kompartmentu 0 dochazi ke konstantni appearance s parametrem 4 (_ --(p4)--> C0)

*/

double CGenerated123_Model::eq_dC0(const double _T, const double _X) const {
	return
		-mState.C[0] * mParameters.vector[2] // A)
		-mState.C[0] * mParameters.vector[3] // B)
		+mParameters.vector[4] // D)
		;
}

double CGenerated123_Model::eq_dC1(const double _T, const double _X) const {
	return
		+mState.C[0] * mParameters.vector[2] // A)
		-(mState.C[1] - mState.C[3]) * mParameters.vector[0] // C)
		;
}

double CGenerated123_Model::eq_dC2(const double _T, const double _X) const {
	return
		+mState.C[0] * mParameters.vector[3] // B)
		-mState.C[2] * mParameters.vector[1] // D)
		;
}

double CGenerated123_Model::eq_dC3(const double _T, const double _X) const {
	return 
		+(mState.C[1] - mState.C[3]) * mParameters.vector[0] // C)
		;
}

void CGenerated123_Model::Emit_All_Signals(double time_advance_delta)
{
	const double _T = mState.lastTime + time_advance_delta;

	// BP NOTE: tohle generujte podle toho, jaky kompartment uzivatel oznacil jako vystupni - parujte signal s kompartmentem
	// tady na prikladu: vystupni signal 1 je generovat z kompartmentu s indexem 2
	Emit_Signal_Level(generated123_model::signal_Output_1, _T, mState.C[2]);
}

HRESULT CGenerated123_Model::Do_Execute(scgms::UDevice_Event event) {
	HRESULT res = S_FALSE;

	if (mInitialized) {

		if (event.event_code() == scgms::NDevice_Event_Code::Level && event.device_time() >= mState.lastTime)
		{
			if (event.signal_id() == generated123_model::signal_Input_1)
			{
				// BP NOTE: jestli uzivatel zadal vstupni signal, musel jej navazat na kompartment
				// tohle propojuje kompartment se vstupnim signalem

				// Na prikladu: vstupni signal 1 (viz podminka nahore) je propojen s kompartmentem 3
				mState.C[3] += event.level();

				res = S_OK;
			}
		}
	}

	if (res == S_FALSE) {
		res = mOutput.Send(event);
	}

	return res;
}

// BP NOTE: nechte jak je
HRESULT CGenerated123_Model::Do_Configure(scgms::SFilter_Configuration configuration, refcnt::Swstr_list& error_description) {
	return S_OK;
}

// BP NOTE: nechte jak je
HRESULT IfaceCalling CGenerated123_Model::Step(const double time_advance_delta) {

	HRESULT rc = E_INVALIDARG;
	if (!mInitialized) {
		return E_ILLEGAL_METHOD_CALL;
	}

	if (time_advance_delta > 0.0) {
		constexpr size_t microStepCount = 5;
		const double microStepSize = time_advance_delta / static_cast<double>(microStepCount);

		for (size_t i = 0; i < microStepCount; i++)
		{
			const double nowTime = mState.lastTime + static_cast<double>(i)*microStepSize;

			for (auto& binding : mEquation_Binding)
				binding.x = ODE_Solver.Step(binding.fnc, nowTime / scgms::One_Minute, binding.x, microStepSize / scgms::One_Minute);
		}

		Emit_All_Signals(time_advance_delta);

		mState.lastTime += time_advance_delta;

		rc = S_OK;
	}
	else if (time_advance_delta == 0.0) {
		Emit_All_Signals(time_advance_delta);
		rc = S_OK;
	}

	return rc;
}

HRESULT CGenerated123_Model::Emit_Signal_Level(const GUID& signal_id, double device_time, double level) {
	scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };

	// BP NOTE: jen generujte spravne jmeno namespace
	evt.device_id() = generated123_model::model_id;
	evt.device_time() = device_time;
	evt.level() = level;
	evt.signal_id() = signal_id;
	evt.segment_id() = mSegment_Id;

	return mOutput.Send(evt);
}

// BP NOTE: nechte jak je
HRESULT IfaceCalling CGenerated123_Model::Initialize(const double current_time, const uint64_t segment_id) {

	if (mInitialized) {
		return E_ILLEGAL_STATE_CHANGE;
	}

	mState.lastTime = current_time;
	mSegment_Id = segment_id;
	mInitialized = true;
	return S_OK;
}
